
#include <sys/types.h>
#include <sys/sysmp.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <cpuset.h>
#include <errno.h>

#define STRLEN  1024

main(int argc, char *argv[])
{
  int i, ierr, status, comstrlen;
  int ncpu    = 1;
  int nset    = 1;
  int verbose = 0;
  int debug   = 0;

  char *usage = "%s: [-v] [-n #cpu]\n";
  char *hostname = NULL;
  char *filename = NULL;
  FILE *permfile;

  pid_t pid;
  uid_t user,  euser;
  gid_t group, egroup;

  cpuset_QueueDef_t *qdef = NULL;
  char *qname   = "runset";
#ifdef DEBUG
  char *command = "./test_script.sh";
#else
  char *command = "/cmiss1/cmiss/examples/common/cmiss_type1/test_tree.pl -s irix64 -p cm_speedup -z 1";
#endif
  char *exe       = "/usr/sbin/cpuset";
  char *comstring = NULL;
  char *format    = NULL;



  user   = getuid();
  group  = getgid();
  euser  = geteuid();
  egroup = getegid();

  /* Read the arguments */

  if (argc > 5 ) { /* Prevent any buffer problems */
    fprintf(stderr,usage,argv[0]);
    exit(0);
  }

  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i],"-v") == 0) {
      verbose = 1;
    }
    else if (strcmp(argv[i],"-d") == 0) {
      debug = 1;
    }
    else if (strcmp(argv[i],"-help") == 0 ||
      strcmp(argv[i],"-h") == 0 ||
      strcmp(argv[i],"-?") == 0) {
      fprintf(stderr,usage,argv[0]);
      exit(0);
    }
    else if (strcmp(argv[i],"-n") == 0 && i < argc-1) {
      if (sscanf(argv[++i],"%d",&nset) != 1) {
        fprintf(stderr,usage,argv[0]);
        exit(0);
      }
    } else {
      fprintf(stderr,usage,argv[0]);
      exit(0);
    }
  }

  /* Get some system info, check the set variables, and open
     the cpuset file */

  if (!(hostname = calloc(MAXHOSTNAMELEN,sizeof(char)))) {
    fprintf(stderr,"%s: out of memory\n",argv[0]);
    exit(0);
  }
  if (!(comstring = calloc(STRLEN,sizeof(char)))) {
    fprintf(stderr,"%s: out of memory\n",argv[0]);
    exit(0);
  }
  gethostname(hostname,MAXHOSTNAMELEN);
  ncpu = cpusetGetCPUCount();
  format = "%s -q %s -A %s";
  comstrlen = strlen(format) + strlen(exe) + strlen(qname) + strlen(command);
  if (comstrlen > STRLEN) {
    fprintf(stderr,"%s: buffer overflow. Must increase STRLEN\n",argv[0]);
    exit(-1);
  }
  sprintf(comstring,"%s -q %s -A %s",exe,qname,command); 

  if (verbose) {
    printf("%s: starting cpu set \"%s\" on host \"%s\" "
           "using %d processors\n",argv[0],qname,hostname,ncpu);
    printf("%s: running comand \"%s\"\n",argv[0],command);
  }

  /* Otherwise we are starting a cpuset: check number of CPU's, and
     create a permissions file. */

  if (nset >= ncpu) {
    fprintf(stderr,"%s: number of cpu's in set >= number of cpu's of host\n",
         argv[0]);
    exit(1);
  }
  if (nset < 1) {
    fprintf(stderr,"%s: number of cpu's in set < 1!\n",argv[0]);
    exit(1);
  }

  filename = tmpnam(NULL);
  if (!(permfile = fopen(filename,"w"))) {
    perror("fopen");
    exit(1);
  }
  if (chmod(filename,S_IRUSR|S_IWUSR|S_IXUSR)) {
    perror("chmod");
    if (!debug) unlink(filename);
    exit(1);
  }
  if (chown(filename,user,group)) {
    perror("chown");
    if (!debug) unlink(filename);
    exit(1);
  }
  if (verbose) {
    printf("%s: using permissions file \"%s\"\n",argv[0],filename);
  }

  /* Create the CPU set */

  if (verbose) printf("%s: Creating cpu set definition\n", argv[0]);
  qdef = cpusetAllocQueueDef(nset);
  if (!qdef) {
    fprintf(stderr,"%s: error allocating cpu set structure\n",argv[0]);
    if (!debug) unlink(filename);
    exit(1);
  }

  /* Define attributes of the cpuset */

  qdef->flags      = CPUSET_CPU_EXCLUSIVE;
  qdef->permfile   = filename;
  qdef->cpu->count = nset;
  for (i = 0; i < nset; i++) {
    qdef->cpu->list[i] = ncpu - ( i + 1 );
    if (verbose) printf("\trequesting cpu %d\n",qdef->cpu->list[i]);
  }

  /* Request that the cpuset be created */

  if (verbose) printf("%s: Creating cpu set \"%s\"\n", argv[0], qname);
  if (!cpusetCreate(qname, qdef)) {
    fprintf(stderr,"%s: error creating cpu set \"%s\"\n",argv[0], qname);
    if (!debug) unlink(filename);
    exit(1);
  }

  /* Run the job */

  if ((pid = fork()) == -1) {
    fprintf(stderr,"%s: forking error\n",argv[0]);
    perror(argv[0]);
    if (!debug) unlink(filename);
    exit(1);
  }

  /* Child process */

  if (pid == 0) {
    if (setuid(user) || setgid(group)) {
      fprintf(stderr,"%s: error su'ing to %d:%d\n",argv[0],user,group);
      perror(argv[0]);
      if (!debug) unlink(filename);
      exit(1);
    }

    if (ierr = system(comstring)) {
      fprintf(stderr,"%s: command \"%s\" returns %d\n",argv[0],comstring,ierr);
    }

    exit(0);
  }

  /* Parent process */

  else {
    wait(&status);
    if (status) {
      fprintf(stderr,"%s: child returns %d\n",argv[0],status);
      perror(argv[0]);
    }
  }

  /* Stop the cpu set */

  if (verbose) {
    printf("%s: stopping cpu set \"%s\" on host \"%s\"\n",
           argv[0],qname,hostname);
  }
  if (setuid(euser) || setgid(egroup)) {
  	fprintf(stderr,"%s: error su'ing to %d:%d\n",argv[0],euser,egroup);
    perror(argv[0]);
    if (!debug) unlink(filename);
    exit(1);
  }
  if (!cpusetDetachAll(qname)) {
    fprintf(stderr,"%s: error detaching processes from cpu set \"%s\"\n",
         argv[0], qname);
    if (!debug) unlink(filename);
    exit(1);
  }
  if (!cpusetDestroy(qname)) {
    fprintf(stderr,"%s: error destroying cpu set \"%s\"\n", argv[0], qname);
    if (!debug) unlink(filename);
    exit(1);
  }

  /* Cleanup */

  cpusetFreeQueueDef(qdef);
  fclose(permfile);
  if (!debug) unlink(filename);

  exit(0);
}
