package test_cmiss;

use strict;

use File::Copy;
use File::Path;

use vars qw($test_summary $test_log);
$test_summary = "test.summary";
$test_log = "log";

my $readdir_grep_string = q/^(\.|CVS|cvs)/;
my $diff_options_ext = '.opt';
my $len_diff_options_ext = length $diff_options_ext;

my %lock_list;

sub find_executable_in_path ( $@ );
sub is_leaf_directory ( $ );

sub sh_quote (@) {
    # quotes and joins arguments for input to shell.
    join ' ', map {
	my $arg = $_;
	# use quotes if arg is zero length or if there are meta characters
	if( $arg eq '' || $arg =~ m'[^-_.:/=A-Z0-9]'i ) {
	    # KAT I can't find a consistent method for quoting
	    # newlines that works in sh and csh.  We don't know what
	    # the (login) shell (that ssh uses) is so abort on
	    # newlines.
	    $arg =~ m/\n/ and
		die "don't know how to quote newlines for shell";
	    $arg =~ s/'/'\\''/g; # quote quotes
	    $arg = "'$arg'"; # quote result
	}
	$arg;
    } @_;
}

sub time_message ($) {
    my @localtime = localtime $_[0];
    # formats the time supplied by gmtime or localtime as HH:MM:SS
    my $sec = $localtime[0];
    my $min = $localtime[1];
    my $hour = $localtime[2];
    sprintf( "%.2d:%.2d:%.2d", $hour, $min, $sec );
}

sub start_lock ($$)
  {
    # The Linux man page for open says that O_EXCL is broken for nfs.
    # Therefore use a lock directory as in cvs.
     my $result_lock_dir = shift;
     my $result_string = shift;

     my $got_lock = \$lock_list{$result_lock_dir};
     if( $$got_lock )
	{
	  die "I already have a lock for $result_lock_dir";
	}

     {
       # Ensure that there is a parent directory in which to make the lock.
       my $parent = $result_lock_dir;
       $parent =~ s|/[^/]*$||;
       -d $parent or mkpath $parent, 0 or warn "can't mkpath $parent: $!";
     }

     # Make sure $result_string contains no slashes as we make a file from it.
     $result_string =~ s|/|!|g;
     # open (5.8.4) also chops trailing \n.
     $result_string =~ s|\n$||g;

     my $last_lock_details = 'none';
     my $sleep_time = 0;

     while( 1 )
       {
	 unless( -d $result_lock_dir )
	   {
	     # Record that we don't know whether we have a lock or not.
	     $$got_lock = '';
	     # Try to create a lock
	     if( mkdir $result_lock_dir, 0755 )
	       {
		 $$got_lock = $result_string;
		 my $filename = "$result_lock_dir/$result_string";
		 open FH, ">$filename" or do
		   {
		     $$got_lock = ''; # prevent further warnings
		     die "can't open $filename: $!";
		   };
		 close FH or die "error closing $filename: $!";
		 last;
	       }
	     else
	       {
		 undef $$got_lock;
		 require POSIX;
		 $! == &POSIX::EEXIST or
		   die "can't create $result_lock_dir: $!\n";
	       }
	   }

	 my $lock_details;
	 if( opendir DIR, $result_lock_dir )
	   {
	     ($lock_details) = ((grep { ! /^(\.\.?)$/ } readdir DIR), "no information file present");
	     closedir DIR;
	   }
	 else
	   {
	     $lock_details = "can't read directory: $!";
	   }


	 if( $lock_details ne $last_lock_details )
	   {
	     print STDERR ("$0:\n [", time_message time,
			   "] Lock $result_lock_dir found, waiting.\n",
			   "$lock_details\n");
	     $sleep_time = 1;
	     $last_lock_details = $lock_details;
	   }
	 else
	   {
	     print STDERR '.';
	     $sleep_time++;
	   }
	 sleep $sleep_time;	#seconds
       }

    if( $sleep_time ) { print STDERR "got lock\n" }
  }

sub lock_output_name ($)
  {
    my ($output_fullpath) = @_;

    # Make lock hidden so that it doesn't look like an output directory.
    $output_fullpath =~ s|^(.*/)([^/]*)$|$1.$2.lock|;
    return $output_fullpath;
  }

sub start_lock_output ($)
  {
    my ($output_fullpath) = @_;

    my $login = getlogin || 'unknown';
    my $program = $0; $program =~ s|^.*/||;
    start_lock( lock_output_name $output_fullpath,
		"$login.$$." . &hostname() . ":$program");
  }

sub end_lock ($)
  {
     my $result_lock_dir = shift;
     #Remove the information file
     my $filename = "$result_lock_dir/$lock_list{$result_lock_dir}";
     unlink $filename or do
       {
	 $lock_list{$result_lock_dir} = ''; # prevent further warnings
	 die "can't remove lock file $filename: $!";
       };
     #Remove the lock directory
     rmdir $result_lock_dir or do
       {
	 delete $lock_list{$result_lock_dir}; # prevent further warnings
	 die "can't remove lock dir $result_lock_dir: $!\n";
       };

     delete $lock_list{$result_lock_dir};
  }

sub end_lock_output ($)
  {
    my ($output_fullpath) = @_;
    end_lock(  lock_output_name $output_fullpath );
  }

sub abort_locks
  {
     foreach my $lockdir ( grep { defined $lock_list{$_} } keys %lock_list )
       {
	 # Make sure no locks are left around.  If the file is '' we really
	 # don't know if we have a lock or not.  But the rmdir will fail if
	 # the locker has created a file.

	 #Don't die on failure, we are already in the abort phase.
	 if( my $file = $lock_list{$lockdir} )
	   {
	     my $filename = "$lockdir/$file";
	     unlink $filename or warn "unable to remove lock file $filename: $!\n";
	   }
	 rmdir $lockdir or warn "unable to remove lock dir $lockdir: $!\n";
       }
  }

sub compare_files
  {
    my $example_dir = shift;
    my $compare_versions_string = shift;
    my $diff_program_paths = shift;
    my $wait_status = shift;
    my $verbose = shift;
    my @diff_types;
    my @passed_output_files = ();
    my @failed_output_files = ();
    
    # Compare files in testing/answers, generic defines what to examine.
    if ( ! -d "$example_dir/testing" ) {
	warn "no testing directory for test example $example_dir";
    } else {
	# Each directory at this level defines a testing type
	opendir(DIR, "$example_dir/testing")
	  || die "can't opendir $example_dir/testing: $!";
	@diff_types = grep { ! /$readdir_grep_string/ } readdir(DIR);
	closedir DIR;
	
	my %missing_files;

	foreach my $diff_program (@diff_types)
	  {
	    my $diff_program_fullname =
	        find_executable_in_path $diff_program, @$diff_program_paths;

	    defined $diff_program_fullname or do {
		print "Diff program $diff_program not found in @$diff_program_paths\n";
		$diff_program_fullname =
		    find_executable_in_path $diff_program,split ':',$ENV{PATH};
	    };

	    if( defined $diff_program_fullname )
	      {
	      my @directory_diff_options = ();
	      if ($diff_program_fullname =~ m/(ndiff|idiff)$/)
		 {
		    my $normal_diff_program = 
		      find_executable_in_path "diff", @$diff_program_paths;
		    # Add the special option of the diff program to use with ndiff
		    if (defined $normal_diff_program)
		      {
			 push (@directory_diff_options, "--diff=$normal_diff_program");
		      }
		 }
	      if ($diff_program_fullname =~ m/idiff$/)
		 {
		    my $imagemagick_dir = 
		      find_executable_in_path "composite", @$diff_program_paths;
		    # Add the special option of the diff program to use with ndiff
		    if (defined $imagemagick_dir &&
		       ($imagemagick_dir =~ m/(.*)composite$/))
		    {
		      $imagemagick_dir = $1;
		      push (@directory_diff_options, "--imagemagick_dir=$imagemagick_dir");
		    }
		 }

	      my %answer_versions = ();
	      # Search the directories in compare_versions_string and
	      # extract the first occurrance of each filename, recording
	      # the directory it is found in.  The filenames are the keys
	      # and the directories are the values of %answer_versions.
	      foreach my $version (split (",", $compare_versions_string)) {
		 my $dir = "$example_dir/testing/$diff_program/$version";
		 if (-d $dir)
		    {
		      opendir(DIR, $dir) || die "can't opendir $dir: $!";
		      # List all files except for hidden ones.
		      my @files =
			 grep { ! /$readdir_grep_string/ } readdir(DIR);
			      closedir DIR;
		      foreach my $file (@files) {
			 unless (exists $answer_versions{$file}) {
			    $answer_versions{$file} = $version;
			 }
		      }
		    }
	      }

	      # Some of the files in %answer_versions are actually
	      # files containing options for the diff program.  If
	      # the file extension indicates an options file and a
	      # corresponding answer exists then remove the options
	      # file from %answer_version and store it as a value of
	      # %options_files under the key of answer.
	      my %options_files = ();
	      foreach my $file (keys %answer_versions) {
		 if( substr( $file, -$len_diff_options_ext)
		      eq $diff_options_ext ) {
		    my $answer = substr $file, 0, -$len_diff_options_ext;
		    if (exists $answer_versions{$answer}) {
		      $options_files{$answer} =
			 "$answer_versions{$file}/$file";
		      delete $answer_versions{$file};
		    }
		 }
	      }
	      
	      while (my ($answer,$answer_version) = each %answer_versions)
		 {
		    #print "Answer $answer\n";
		    if (-f $answer)  # Has the file been generated?
		      {
			 my @diff_options = @directory_diff_options;
			 if (exists $options_files{$answer}) {
			    my $file = "$example_dir/testing/$diff_program/$options_files{$answer}";
			    open (OPTFILE, "<$file") ||
			      die "can't open $file: $!\n";
			    push (@diff_options, <OPTFILE>);
			    close OPTFILE;
			 }
			 chomp @diff_options;
			 my $correct_answer =
"$example_dir/testing/$diff_program/$answer_version/$answer";
			 my @command =
			    ($diff_program_fullname, @diff_options,
			     $correct_answer, $answer);

			 #Debug 
			 if ($verbose)
			   {
			     print sh_quote (@command) . "\n";
			   }
			 # Write the file header, including a dummy location
			 # for status which is filled in below
			 my $difffile = "$answer.$diff_program";
			 open (DIFFFILE, ">$difffile") || die "can't open $difffile: $!\n";
			 # copy all permissions from answer except user write
			 &copy_permissions(07577,$correct_answer,$difffile);
			 print DIFFFILE "***** CMISS_testing $answer $diff_program $answer_version ";
			 my $status_index = tell DIFFFILE;
			 # *Status is exactly 7 letters as it will be overwritten
			 # with either Success or Failure below.
			 print DIFFFILE "*STATUS *****\n";

			 # redirect STDOUT, STDERR to DIFFFILE, and run command.
			 # (what happens if we recieve and interrupt here?)
			 open(OLDOUT, ">&STDOUT") ||
			    die "Can't dup STDOUT: $!\n";
			 open(OLDERR, ">&STDERR") ||
			    die "Can't dup STDERR: $!\n";
			 open(STDOUT, '>&DIFFFILE') ||
			    die "Can't dup DIFFFILE: $!\n";
			 open(STDERR, '>&DIFFFILE')  ||
			    die "Can't dup DIFFFILE: $!\n";
			 close DIFFFILE; # flush output
			 
			 my $result_code = system @command;
			 
			 open(DIFFFILE, ">&STDOUT") ||
			    die "Can't dup STDOUT: $!\n";		
			 open(STDOUT, ">&OLDOUT") ||
			    die "Can't recover STDOUT: $!\n";
			 open(STDERR, ">&OLDERR") ||
			    die "Can't recover STDERR: $!\n";
			 close OLDOUT;
			 close OLDERR;
			 
			 # Overwrite the status tag in the header. 7 letters.
			 seek (DIFFFILE, $status_index, 0);
			 if ($result_code)
			    {
			      push (@failed_output_files, "$answer.$diff_program");
			      print DIFFFILE "Failure"; # 7 letters
			    }
			 else
			 {
			     push (@passed_output_files, "$answer.$diff_program");
			     print DIFFFILE "Success"; # 7 letters
			 }
			 close DIFFFILE or die "error closing $difffile: $!\n";
		     }
		    elsif( ! exists $missing_files{$answer} )
		    {
			$missing_files{$answer} = 1;

			my $diff_file = "$answer.missing";
			open DIFFFILE, ">$diff_file" or
			      die "can't open $diff_file: $!\n";
			print DIFFFILE "***** CMISS_testing $answer $diff_program $answer_version ";
			print DIFFFILE "Missing *****\n";
			print DIFFFILE "Required test file $answer not created\n";
			close DIFFFILE or die "error closing $diff_file: $!\n";

			push (@failed_output_files, $diff_file);
		     }
		 }
	      }
	    else
	      {
		  die "diff program $diff_program not found\n";
	      }
	  }
    }

    if( my $signal_num  = $wait_status & 0x7f ) {
	my $signal_name = 'SIG'.&sig_name($signal_num);
	my $file = "$signal_name.exit";
	open DIFFFILE, ">$file" or die "can't open $file\n";
	print DIFFFILE
	    "***** CMISS_testing signal exit  $signal_name *****\n";
	close DIFFFILE or die "error closing $file: $!\n";
	push @failed_output_files, $file;
    }
    if( my $exit_status = $wait_status >> 8 ) {
	my $file = "status_$exit_status.exit";
	open DIFFFILE, ">$file" or die "can't open $file\n";
	print DIFFFILE
	    "***** CMISS_testing status exit  $exit_status *****\n";
	close DIFFFILE or die "error closing $file: $!\n";
	push @failed_output_files, $file;
    }

    return(\@failed_output_files, \@passed_output_files);
  }

sub abort
  {
    # Die but make sure we don't leave the lock file lying around.
    test_cmiss::abort_locks();
    die "$0: ",@_;
  }

sub recurse_directory_tree ( &&$@ ) {
      my ($directory_enter_function,$directory_exit_function,
	  $path,@data) = @_;

      # Traverses through a directory tree executing $directory_enter_function
      # on entry to and $directory_exit_function on exit from each
      # directory.  $directory_enter_function may modify (sort, prune, etc.)
      # the list of subdirectories to descended.
      # Returns the return code of the exit function called
      # at the top level.

      # $branch_enter_function is called with the arguments
      # $path,\@subdirs,\@files,@data
      # where @files and @subdirs are the non-directory and directory contents
      # of the current directory,
      # $branch_exit_function is called with arguments
      # $path,\@results,\@files,@data
      # where @files are the non-directory files of the current directory.
      # where @results contains the results from the descended subdirectories.

      opendir DIR, $path or
	die "can't opendir $path: $!\n";

      my @contents = grep { ! /^(\.\.?)$/o } readdir DIR;

      closedir DIR;

      my (@files,@subdirs);

      foreach my $item (@contents)
	{
	  # don't include symbolic links in directories.
	  lstat "$path/$item";
	  if( -d _ )
	    { push @subdirs, $item }
	  else
	    { push @files, $item }
	}

      undef @contents;

      my @results;

      $directory_enter_function->($path,\@subdirs,\@files,@data);
      #Look further down.

      @results =
	map { my $dir = $path eq '.' ? $_ : "$path/$_";
	      recurse_directory_tree( $directory_enter_function,
				      $directory_exit_function,
				      $dir, @data )
	    } @subdirs;

      return $directory_exit_function->($path,\@results,\@files,@data);
    }

sub recurse_example_tree ( &&&$$@ ) {
    my ($leaf_function,$branch_enter_function,$branch_exit_function,
	$examples_root,$subpath,@data) = @_;

    # Traverses through the example tree executing
    # $branch_enter_function on entry to each non-leaf node,
    # $leaf_function on each leaf, and
    # $branch_exit_function on exit from each non-leaf node.
    # If branch_enter_function returns false the subdirectory is not descended.
    # Returns the return code of the leaf or branch-exit function called
    # at the top level.

    # $leaf_function and $branch_enter_function are called with the arguments
    # $examples_root,$subpath,@data.
    # $branch_exit_function is called with arguments
    # $examples_root,$subpath,@return_codes
    # where @return_codes contains the return codes from all the
    # subdirectories.

    my $example_fullpath =
	$subpath eq '' ? $examples_root : "$examples_root/$subpath";

    if( is_leaf_directory $example_fullpath ) {
	return $leaf_function->($examples_root,$subpath,@data);

    } elsif( $branch_enter_function->($examples_root,$subpath,@data) ) {
	#Look further down.

	opendir DIR, $example_fullpath or do {
	    warn "can't opendir $example_fullpath: $!\n";
	    return;
	};

	my @subdirs =
	    sort { $a cmp $b } grep {
		! /$readdir_grep_string|^webdata$/ && -d "$example_fullpath/$_"
		} readdir DIR;

	closedir DIR;

	length $subpath and @subdirs = map { "$subpath/$_" } @subdirs;

	my @return_codes = map {
	    recurse_example_tree( $leaf_function, $branch_enter_function,
				  $branch_exit_function, $examples_root,
				  $_, @data )
	    } @subdirs;

	return $branch_exit_function->($examples_root,$subpath,@return_codes);
    }
}

sub generate_summary
  {
    my $example_dir = shift;
    my $example_path = shift;
    my $compare_version = shift;
    my $result_path = shift;
    my $failed_output_ref = shift;
    my $passed_output_ref = shift;
    my $start_time = shift;
    my $clock_time = shift;
    my $cpu_times_ref = shift;
    my $keep_all_files = shift;
    my $verbose = shift;
    # I would use test.mail but the .mail extension is associated with
    # a mail mime type.
#     my $webpath = ();
#     if ($ENV{CMISS_WWW_URL})
#     {
# 	$webpath = "$ENV{CMISS_WWW_URL}help/examples/example_files";
#     }
#     else
#     {
# 	$webpath = "CMISS_WWW_URL";
#     }

    local %SIG;
    $SIG{__DIE__} = \&abort;
    $SIG{'INT'} = \&abort;

    #Construct the mail file which reports any errors
    open (MAILFILE, ">$test_summary") || die "can't open $test_summary: $!\n";
    if (@$failed_output_ref)
    {
	print MAILFILE "***** CMISS_testing $test_summary test_cmiss.pm $compare_version Failure *****\n";
	print MAILFILE "Clock_execution $clock_time seconds.  CPU times @$cpu_times_ref\n";
#  	if ($ENV{CMISS_WWW_URL})
#  	{
#  	    print MAILFILE "Problems with $webpath/$example_path\n";
#  	}
#  	else
#  	{
#  	    print MAILFILE "Problems in test for example $example_path\n";
#  	}
	for (@$failed_output_ref)
	{
	    print MAILFILE "File $_ failed\n";
	}
    }
    else
    {
	print MAILFILE "***** CMISS_testing $test_summary test_cmiss.pm $compare_version Success *****\n";
	print MAILFILE "Clock_execution $clock_time seconds.  CPU times @$cpu_times_ref\n";
    }
    close MAILFILE or die "error closing $test_summary: $!\n";

    #If there is a result path then copy all the diff_output_files to
    #that location
    #If there is not a result path or we are in verbose mode cat the MAILFILE to the screen.
    if ($verbose || !$result_path)
      {
	print "\n";
	open (MAILFILE, "<$test_summary") || die "can't open $test_summary:	 $!\n";
	print <MAILFILE>;
	close MAILFILE;
      }

    if ($result_path)
    {
	unless (-d $result_path)
	{
	    mkpath($result_path, 0, 0755) or
		die "can't mkdir $result_path: $!\n";
	}

	start_lock_output $result_path;

	#Compare directories to determine if results have changed.

	#assume same results as yesterday until a different result is found
	my $different_failures = 0;
	foreach my $new_file ( @$failed_output_ref ) {
	    my $old_file = "$result_path/$new_file";
	    if( ! -f $old_file ) {
		$different_failures = 1;
	    } elsif( $new_file =~ /\.n?diff$/ ) {
		# Need to ignore time stamps so we do a special
		# comparison ourselves.
		open NEW_FILE, $new_file or
		    die "can't open $new_file: $!\n";
		open OLD_FILE, $old_file or
		    die "can't open $old_file: $!\n";
		for(my $line_num = 0; 1; $line_num++) {
		    my $new_line = <NEW_FILE>;
		    my $old_line = <OLD_FILE>;
		    defined $new_line or defined $old_line or last;
		    # line 1 and 2 contain timestamps so ignore these
		    if( ! defined $new_line or ! defined $old_line or
			( $new_line ne $old_line and
			  $line_num != 1 ||
			  $old_line !~ /^--- / || $new_line !~ /^--- / and
			  $line_num != 2 ||
			  $old_line !~ /^\+\+\+ / || $new_line !~ /^\+\+\+ / ))
		    {
			$different_failures = 1;
			last;
		    }
		}
		close NEW_FILE;
		close OLD_FILE;
	    } else {
		my @command = ('cmp', '-s', $old_file, $new_file);
		if( system @command ) {
		    $? == -1 and die "error executing $command[0]: $!\n";
		    $? != 0x100 and
			die "@command returned ", $? >> 8, "\n";
		    $different_failures = 1;
		}
	    }
	    $different_failures and last;
	}

	# Record the run in the log file before removing old results
	# in case we are interupted.

	my $log_file = "$result_path/$test_log";
	my $new_log = ! -f $log_file;
	open LOG, ">>$log_file" or die "can't open >>$log_file: $!\n";
	if( $new_log ) {
	    print LOG "start time       status  host  real  user  sys\n";
	}
	my $status = do {
	    if( $different_failures ) {
		'break';
	    } elsif( @$failed_output_ref ) {
		'broken';
	    } else {
		'success';
	    }
	};
	printf( LOG "%s %-7s %-5s %4d %5.2f %5.2f\n",
		&time_stamp($start_time), $status, &hostname(),
		$clock_time, @$cpu_times_ref );
	close LOG  or die "error closing $log_file: $!\n";

        #Empty the directory if any files are in it.
        if (!(opendir(DIR, "$result_path")))
	{
	    die("can't opendir $result_path: $!");
	}
        my @result_dir =
	    grep { ! /$readdir_grep_string/ && $_ ne $test_log } readdir(DIR);
        closedir DIR;
        foreach my $file (@result_dir)
	{
	    my $full_name = "$result_path/$file";
	    unlink $full_name or die "can't unlink $full_name: $!\n";
	}

	for ($test_summary,@$failed_output_ref,@$passed_output_ref)
	{
	    move $_, $result_path or
		die "can't move $_ to $result_path: $!\n";
	}

	if ($keep_all_files)
	{
	    #Keep all the files in the testing directory.
	    if (!(opendir(DIR, ".")))
	    {
		die("can't opendir current directory: $!\n");
	    }
	    my @current_dir = grep { ! /$readdir_grep_string/ } readdir(DIR);
	    closedir DIR;
	    for (@current_dir)
	    {
		# rename files called $test_log so they don't overwrite the log
		s/^($test_log)$/$1.retain/;
		move ($_, "$result_path/$_");
	    }
	}

	end_lock_output $result_path;
      }
  }

sub last_break_and_success( $ ) {
    my ($result_path) = @_;

    # Return info for the last break, last success, and last test for testing
    # corresponding to the given result path. The result is a reference to a
    # hash with keys success, break, last, and prebreak, where prebreak
    # indicates the last test before the break.  The values of the hash are
    # references to hashes with keys 'start time' and 'real time'.

    my %info;
    my ($pre_info);
    my $log_file = "$result_path/$test_log";
    if( -f $log_file ) {
	open LOG, $log_file or die "can't open $log_file: $!\n";
	scalar <LOG>; # skip header line
	while( defined( my $line = <LOG> ) ) {
	    my ($time_stamp,$status,$machine,$clock_time,$user_cpu,$sys_cpu) =
		split ' ', $line;

	    my $info = { time_stamp => $time_stamp,
			 'real time' => $clock_time };
	    if( $status eq 'success' ) {
		$info{success} = $info;
		delete $info{break};
		delete $info{prebreak};
	    } elsif( $status eq 'break' ) {
		$info{break} = $info;
		$info{prebreak} = $pre_info;
	    }
	    $pre_info = $info;
	}
	close LOG;

	if( defined $pre_info ) # There has been a test
	  { # Record the last test
	    $info{last} = $pre_info;
	  }

	foreach my $info (values %info) {
	    if( defined $info && exists $info->{time_stamp}) {
		$info->{'start time'} =
		    &interpret_time_stamp($info->{time_stamp});
		delete $info->{time_stamp};
	    }
	}
    }

    return \%info;
}

#I have changed getopts from Getopt::Std so that it just ignores unknown options, I also always use it with a hash.
sub getopts ($;$) {
  my $argumentative;
  my $hash;
  ($argumentative, $hash) = @_;
  my @args;
  my $first;
  my $rest;
  local($_);
  my $errs = 0;
  my @EXPORT;
  my $pos;

    @args = split( / */, $argumentative );
    while(@ARGV) {
      if (($_ = $ARGV[0]) =~ /^-(.)(.*)/) {
        ($first,$rest) = ($1,$2);
        if (/^--$/) {   # early exit if --
            shift @ARGV;
            last;
        }
        $pos = index($argumentative,$first);
        if ($pos >= 0) {
            if (defined($args[$pos+1]) and ($args[$pos+1] eq ':')) {
                shift(@ARGV);
                if ($rest eq '') {
                    ++$errs unless @ARGV;
                    $rest = shift(@ARGV);
                }
                if (ref $hash) {
                    $$hash{$first} = $rest;
                }
                else {
                    ${"opt_$first"} = $rest;
                    push( @EXPORT, "\$opt_$first" );
                }
            }
            else {
                if (ref $hash) {
                    $$hash{$first} = 1;
                }
                else {
                    ${"opt_$first"} = 1;
                    push( @EXPORT, "\$opt_$first" );
                }
                if ($rest eq '') {
                    shift(@ARGV);
                }
                else {
                    $ARGV[0] = "-$rest";
                }
            }
        }
        else {
	    #SAB I don't want this to be an error, just continue.
            #warn "Unknown option: $first\n";
            #++$errs;
	    #KAT: The option hasn't been recognized.
	    #     $rest may be a value for the argument so
	    #     don't split up rest into single letter options.
#              if ($rest ne '') {
#                  $ARGV[0] = "-$rest";
#              }
#              else {
#                  shift(@ARGV);
#              }
	    shift(@ARGV);
        }
      }
      else {
	shift(@ARGV);
      }
    }
    unless (ref $hash) { 
        local $Exporter::ExportLevel = 1;
        #import Getopt::Std;
    }
    $errs == 0;
}

sub read_notify_list
  {
    my $notify_list_filename = shift;
    my $notify_example_ref = shift;
    my $filename;
    
    my $notify_list_file = new FileHandle("<$notify_list_filename")
      || die "can't open $notify_list_filename: $!\n";
    
    while (<$notify_list_file>)
      {
	if (m"^!")
	  {
	    #Comment, do nothing
	  }
	elsif (m"^\s*$")
	  {
	    #Skip blank lines
	  }
	elsif (m"^#")
	  {
	    #Directive
	    if (m'^#include "(.*)"')
	      {
		$filename = $1;
		$filename =~ s%\$[(](.*?)[)]%$ENV{$1}%g;
		read_notify_list($filename, $notify_example_ref);
	      }
	    else
	      {
		die "Unknown directive \"$_\" in file $notify_list_filename";
	      }
	  }
	else
	  {
	    #Assume this is a mail address or a , separated list.
	    #Commas are used in the batch mail lists already so 
	    #it is useful to be able to just use the same lists.
	    chomp $_;
	    push (@$notify_example_ref, split (",", $_));
	  }
      }
    close ($notify_list_file);
  }

#Recurse down tree.
sub make_mail_trees {
    my ($output_location,$example_location,$example_paths_ref,
		$programmer_name_regexp,$notify_all_ref) = @_;

    # output variables
    my %notify_hash = ();
    my @failure_list = ();

    foreach my $example_path (@$example_paths_ref) {
	my ($output_dir,$example_dir) = $example_path eq '' ?
	    ($output_location,$example_location) :
		("$output_location/$example_path",
		 "$example_location/$example_path");

	if( ! is_leaf_directory("$example_dir"))
	{
	    #Look further down.

	    opendir(DIR, "$output_dir") or do
	      {
		warn "can't opendir $output_dir: $!";
		next;
	      };
	    my @subdirectories =
		sort {$a cmp $b} grep {
		    ! /$readdir_grep_string/ && -d "$output_dir/$_"
		    } readdir(DIR);
	    closedir DIR;

	    my @example_paths = length ($example_path) > 0 ?
		map { "$example_path/$_" } @subdirectories :
		    @subdirectories;

	    undef @subdirectories;

	    my ($notify_hash_ref,$failure_list_ref) =
		make_mail_trees( $output_location, $example_location,
				 \@example_paths,
				 $programmer_name_regexp, $notify_all_ref);
	    @notify_hash{ keys %$notify_hash_ref } = ();
	    push @failure_list, @$failure_list_ref;
	} elsif( -f "$example_dir/control/test" ) {

	    # Only checking notification and output directories if
	    # there is a test script.  If the test script has been
	    # removed and there are still output directories assume
	    # that the result of the old test are not intended to be
	    # reported.

	    my @notify_example_list = ();
	    if (defined $programmer_name_regexp and
		-f "$example_dir/control/notify_list")
	    {
		read_notify_list("$example_dir/control/notify_list", 
				 \@notify_example_list);
		@notify_example_list =
		    grep {/^$programmer_name_regexp$/} @notify_example_list;
	    }
	    push @notify_example_list, @$notify_all_ref;
	    #record these in the global notify hash.
	    @notify_hash{@notify_example_list} = ();

	    #Only bother to check failure if there is a notify_list.
	    if (@notify_example_list ) {
		#Find example status
		$output_dir = "$output_dir/output";
		if (-d $output_dir)
		{
 		    my @subdirectories;
		    test_cmiss::add_leaf_directories(\@subdirectories,
		       "$output_dir", "");
		    @subdirectories = sort { $a cmp $b } @subdirectories;
		    my @failed_versions;
		    my @break_times;
		    my @incomplete_versions;
		    my $latest_prebreak = 0;
		    foreach my $version (@subdirectories)
		    {
			my $result_path = "$output_dir/$version";
			my $summary_file = "$result_path/$test_summary";
			my ($version_time,$prebreak_time);
			
			if( -f $summary_file ) {
			    open TEST_FILE, $summary_file
				or die "can't open $summary_file: $!\n";
			    my $line;
			    if( ! defined( $line = <TEST_FILE> )
or $line !~ /^\Q*****\E CMISS_testing(?: \S+){3} (\S+) \Q*****\E\n/ )
			    {
				warn "header of $summary_file not recognized\n";
				push (@failed_versions, $version);
				stat $summary_file or
				    die "can't stat $summary_file: $!\n";
				$version_time = (stat _)[9];
			    }
			    else {
				my $status = $1;
				if ($status =~ /^success/i) {
				    #Do nothing
				} elsif( $status =~ /^pending/i ) {
				    push @incomplete_versions, $version;
				    # force a new report
				    $version_time = time;
				    $prebreak_time = $version_time;
				} else {
				    push @failed_versions, $version;
				    my ($break_info,$prebreak_info) = @{
					last_break_and_success $result_path
					}{qw(break prebreak)};
				    $version_time =
					$break_info->{'start time'};
				    defined $prebreak_info and
				    $prebreak_time =
				        $prebreak_info->{'start time'};
				}
			    }
			    close TEST_FILE;
			}
			else
			{
			    push (@failed_versions, $version);
			    $version_time = time;
			}

			# Record the lastest test before a break as a
			# lower limit on the time for the lastest
			# break.
			if( defined $prebreak_time
			    && $prebreak_time > $latest_prebreak ) {
			    $latest_prebreak = $prebreak_time;
			}
			if( defined $version_time ) {
			    push @break_times, $version_time;
			}
		    }
		    if (@failed_versions || @incomplete_versions)
		    {
			my $name;
			if (-f "$example_dir/web_data/name.txt") {
			    open (NAME_FILE, "<$example_dir/web_data/name.txt")
				|| die "can't open $example_dir/web_data/name.txt: $!\n";
			    $name = <NAME_FILE>;
			    chomp $name;
			    close NAME_FILE;
			} else {
			    $name = "";
			}
			# Find the earliest detected break after the
			# last detected prebreak and report this as
			# the time the break was detected.
			my $break_time;
			foreach my $time (@break_times) {
			    if( $time >= $latest_prebreak ) {
				if( ! defined $break_time
				    || $time < $break_time )
				{
				    $break_time = $time;
				}
			    }
			}
			push @failure_list,
		        { 'example path' => $example_path,
			  'name' => $name,
			  'failed versions' => \@failed_versions,
			  'incomplete versions' => \@incomplete_versions,
			  'break time' => $break_time,
			  'notify list' => \@notify_example_list };
		    }
		}
	    }
	}
    }
    return \%notify_hash, \@failure_list;
}

sub is_leaf_directory ( $ )
  {
     #This routine answers the question, Is this a leaf as far as all 
     #the tree recursion routines are concerned?
     #The current criteria is that if a "control" directory exists and
     #contains any non directory files then this is a leaf.  The 
     #exclusion of directories is because these are hard to get rid of
     #reliably with CVS.
     my $example_dir = shift;
     my $return_code;
     my $filename;

     $return_code = 0;
     if (-d "$example_dir/control")
	{
	  opendir(DIR, "$example_dir/control") 
	      || die "can't opendir $example_dir/control: $!";
	  my @control_dir = grep { ! /$readdir_grep_string/ } readdir(DIR);
	  closedir DIR;

	  for $filename (@control_dir)
	     {
		if (-f "$example_dir/control/$filename")
		  {
		     $return_code = 1;
		  }
	     }
	}
     return($return_code);
  }

sub find_executable_in_path ( $@ )
{
    my ($name,@paths) = @_;
# Searches @paths for an executable called $name.  Returns the full filename
# (including path) of the first executable found, or undef if no executable
# is found.  A slash is appended to the directories in @paths.
    foreach my $dir (@paths) {
	my $fullname = "$dir/$name";
	if( -x $fullname ) { return $fullname }
    }
    return; # not found
}

sub hostname () {
    if( defined $ENV{'HOST'} ) {
	return $ENV{'HOST'};
    } else {
	require POSIX;
	my $hostname = (POSIX::uname())[1];
	$hostname =~ s/\.auckland\.ac\.nz$//; #trim domain
	$ENV{'HOST'} = $hostname;
	return $hostname;
    }
}

sub html_link {
    my ($www_url,$location,$file,$text) = @_;
    # Produces html text for a hyperlink to $file (specified as
    # relative to $location) labelled with $text.  If the file
    # is world readable then $www_url is used; otherwise
    # a direct file reference is used.  If the file does not yet exist
    # (e.g. html_tree has not yet been run) then use a file reference
    # rather than dying.

    my $fullname = "$location/$file"; 
    return '<a href="' .
	( stat $fullname && (stat _)[2] & 0004 ?
	  "$www_url$file" : "file://localhost$fullname" ) .
	"\">$text</a>\n";
}

sub copy_permissions {
    my ($set,$from,$to) = @_;
    # Copy the specified set of permissions of $from to those of $to
    # but leave the other permissions unchanged.  If any of the
    # permissions copied are group permissions then the group is also
    # copied.

    my @from_stat = stat $from or die "can't stat $from: $!\n";
    my @to_stat = stat $to or die "can't stat $to: $!\n";
    my $group = $from_stat[5];
    $set & 0070 and
	chown -1, $group, $to or die "can't chgrp $group $to: $!\n";
    my $permissions = 07777 & (($to_stat[2] &~ $set) | ($from_stat[2] & $set));
    chmod $permissions, $to or die "can't chmod $permissions $to: $!\n";
}

sub time_stamp {
    my @gmtime = gmtime $_[0];
#      my $sec = $gmtime[0];
    my $min = $gmtime[1];
    my $hour = $gmtime[2];
    my $mday = $gmtime[3];
    my $mon = $gmtime[4] + 1;
    my $year = $gmtime[5] + 1900;
    sprintf "%.4d-%.2d-%.2d,%.2d:%.2d",$year,$mon,$mday,$hour,$min;
}

sub interpret_time_stamp ($) {
    my ($time_stamp) = @_;
    $time_stamp =~ m/(\d+)-(\d+)-(\d+),(\d+):(\d+)/
      or die "unrecognized time stamp `$time_stamp'\n";
    my $sec = 0;
    my $min = $5;
    my $hour = $4;
    my $mday = $3;
    my $mon = $2 - 1;
    my $year = $1 - 1900;
    require Time::Local;
    return &Time::Local::timegm($sec,$min,$hour,$mday,$mon,$year);
}

{
    my %sig_num;
    my @sig_name;

    sub sig_name {
	my ($signal_num) = @_;
	unless( @sig_name ) {
	    require Config;
	    unless($Config::Config{sig_name} && $Config::Config{sig_num}) {
		die "No sigs?";
	    }
	    my @names = split ' ', $Config::Config{sig_name};
	    @sig_num{@names} = split ' ', $Config::Config{sig_num};
	    foreach my $name (@names) {
		$sig_name[$sig_num{$name}] ||= $name;
	    }
	}
	return $sig_name[$signal_num];
    }
}

sub add_leaf_directories
  {
     my $array_ref = shift;
     my $root_directory = shift;
     my $current_directory = shift;
     my @current_directory_list;
     my $found_subdirectory;
     my $file;


     $found_subdirectory = 0;

     opendir(DIR, "$root_directory/$current_directory") || die "can't opendir $root_directory/$current_directory: $!\n";
     # Get all files except for hidden ones.
     @current_directory_list = grep { ! /$readdir_grep_string/ } readdir(DIR);
     #Make sure we don't recurse while the DIR filehandle is still active as it is global.
     closedir DIR;

     for $file (@current_directory_list)
       {
	 if (-d "$root_directory/$current_directory/$file")
	   {
	     $found_subdirectory = 1;
	     #We need to be a bit special on the root
	     if ($current_directory)
	       {
	          add_leaf_directories($array_ref, "$root_directory", "$current_directory/$file");
	       }
	     else
	       {
	          add_leaf_directories($array_ref, "$root_directory", "$file");
	       }
	   }
       }

     if (!$found_subdirectory || -f "$current_directory/$test_summary")
       {
	 push (@$array_ref, $current_directory);
       }
  }

# The module compile time return
return 1;

### Local Variables: 
### tab-width: 8
### End: 
